# encoding: utf-8

'''Resources and tasks for data pipes'''

from .plugins_ops import PluginCollection
from .utils.misc_utils import merge_dict, merge_dicts, nested_set, collect_dicts, MyException
import copy
import datetime
import networkx as nx
import os
import shlex
import subprocess
import time, sys
from tqdm import tqdm
from jpl.pipedreams.celeryapp import CeleryDreamer, obj_func_runner, indi_func_runner
import inspect

class Resource(object):

    def __init__(self, ID: str, **kwargs):
        self.ID = ID
        self.resources = kwargs


class Task(object):

    def __init__(self, name: str, process, resource_ID: str, plugin_collection: PluginCollection, celerydreamer,
                 run_function=None, params: dict = None, op_params: dict = None):

        if params is None:
            params = {}
        if op_params is None:
            op_params = {}

        is_plugin = False
        if type(process) == str:
            process_name = str(process)
            process = plugin_collection.get_plugin(process)
            is_plugin = True
        elif hasattr(process, '__call__'):
            process_name = process.__name__
        else:
            pass  # todo : throw error
            print('process type not recognized: ' + str(type(process)))
            return

        if run_function is not None:
            process_name += '..' + run_function

        self.name = name
        self.resource_ID = resource_ID
        self.process = process
        self.process_name = process_name
        self.is_plugin = is_plugin
        self.result = {}
        self.params = params
        self.op_params = op_params
        self.run_function = run_function
        self.celerydreamer = celerydreamer
        self.time_taken=None # in seconds
        self.objects={} # these objects will go into the process like a param that can be added upon by the process

    @staticmethod
    def get_process_name(process):
        if type(process) == str:
            return process
        elif hasattr(process, '__call__'):
            return process.__name__
        else:
            MyException('Error: Process name could not be resolved!')

    @staticmethod
    def concoct_task_ID(name, resource_ID):
        return name + '|+|' + resource_ID

    @staticmethod
    def task_ID_to_resource_ID(task_ID):
        return task_ID.split('|+|')[1]

    @staticmethod
    def task_ID_to_function_name(task_ID):
        return task_ID.split('|+|')[0]

    def get_task_ID(self):
        return Task.concoct_task_ID(self.name, self.resource_ID)

    def run_task(self, single_process=True, **kwargs):
        for k, v in self.params.items():
            kwargs[k] = v

        self.params = kwargs

        if self.is_plugin:
            if self.run_function is not None:
                if single_process:
                    self.result = obj_func_runner(self.process, self.run_function, self.objects, **kwargs)
                else:
                    self.result = self.celerydreamer.celery_obj_func_runner.delay(self.process, self.run_function, self.objects, **kwargs)
            else:
                if single_process:
                    self.result = obj_func_runner(self.process, 'run', self.objects, **kwargs)
                else:
                    self.result = self.celerydreamer.celery_obj_func_runner.delay(self.process, 'run', self.objects, **kwargs)
        else:
            if single_process:
                self.result = indi_func_runner(self.process, self.objects, **kwargs)
            else:
                self.result = self.celerydreamer.celery_indi_func_runner.delay(self.process, self.objects, **kwargs)

    def postprocess_results(self):

        # convert the result into a dict based on the op_params if not already a dict!
        if "op_out" in self.op_params.keys():
            if type(self.result) != tuple:
                self.result = tuple([self.result])
            result_dict = {}
            for i, k in enumerate(self.op_params["op_out"]):
                result_dict[k] = self.result[i]
            self.result = result_dict

        # in case the function does not return any results
        if self.result is None:
            self.result = {}

        # a process metadata that tracks which tasks have been applied to the artifact so far
        process_undergone = self.params.get('processes_undergone', list())
        process_undergone.append(self.name)
        self.result['processes_undergone'] = process_undergone


def register_connector_func(func):
    """
    a dummy decorator
    """
    return func


def register_non_parallelizable_connector_func(func):
    """
    a dummy decorator
    """
    return func


def methodsWithDecorator(cls, decoratorName):
    """
    Find all the methods names of a class with a particular decorator applied
    """
    sourcelines = inspect.getsourcelines(cls)[0]
    for i, line in enumerate(sourcelines):
        line = line.strip()
        if line.split('(')[0].strip() == '@' + decoratorName:  # leaving a bit out
            nextLine = sourcelines[i + 1]
            name = nextLine.split('def')[1].split('(')[0].strip()
            yield (name)


class Operation(object):

    # ======== result connection/manipulation help functions:

    @staticmethod
    @register_connector_func
    def change_name(in_to_out: dict, **kwargs):
        return {in_to_out.get(k, k): v for k, v in kwargs.items()}

    @staticmethod
    @register_connector_func
    def remove_keys(keys_to_remove, result_levels=None, **kwargs):
        if result_levels is None:
            result_levels = []
        level_value = kwargs
        for level in result_levels:
            level_value = level_value[level]
        level_value_edit = {k: v for k, v in level_value.items() if k not in keys_to_remove}
        kwargs = nested_set(kwargs, result_levels, level_value_edit)
        return kwargs

    @register_non_parallelizable_connector_func
    def add_to_op_resource(self, in_to_op: dict, **kwargs):
        for resource_name, resource_name_new in in_to_op.items():
            if kwargs['task_ID'] not in self.added_resources.keys():
                self.added_resources[kwargs['task_ID']] = {}
            self.added_resources[kwargs['task_ID']][resource_name_new] = kwargs[resource_name]
            # print('DBG adding 1:', kwargs['task_ID'], resource_name_new, kwargs[resource_name])
            # print('DBG adding 2:', self.added_resources)
        kwargs.pop('task_ID')
        return kwargs

    @staticmethod
    @register_connector_func
    def merge(in_multi, out="collapsed_results", merge_type='forced_ordered', **kwargs):
        list_of_dict = []
        if type(in_multi) == list:
            for result_name in in_multi:
                list_of_dict.append(kwargs.get(result_name, {}))
                kwargs.pop(result_name) if result_name in kwargs.keys() else 0

        if type(in_multi) == str and in_multi in kwargs.keys():
            if out == 'collapsed_results':
                out = in_multi
            list_of_dict = kwargs[in_multi]
        if merge_type == 'forced_ordered':
            kwargs[out] = merge_dicts(list_of_dict)
        elif merge_type == 'collect':
            kwargs[out] = collect_dicts(list_of_dict)
        return kwargs

    @staticmethod
    @register_connector_func
    def artificial_wait(wait_time, **kwargs):
        time.sleep(wait_time)
        return kwargs

    def generate_process_ID(self):
        self.counter+=1
        return str(self.counter)

    def __init__(self, name: str, redis_path=None, include_plugins=None):
        self.instructions = []
        self.task_graph = nx.DiGraph()
        self.name = name
        self.task_ID_to_task = {}
        self.plugin_collection = PluginCollection()
        self.results = {}
        self.inherited_params = {}
        self.self_params = {}
        self.added_resources = {}
        self.inherited_resource = {}
        self.redis_path = redis_path
        self.include_plugins = include_plugins
        self.times = {}
        self.cache = {}
        self.templates={}
        self.counter=0

        # add all the connector functions to a dictionary; to be accessed using their names
        self.connector_functions_mapping = {func_name: getattr(Operation, func_name) for func_name in
                                            list(methodsWithDecorator(Operation, 'register_connector_func'))}
        self.non_parallel_connector_functions_mapping = {func_name: getattr(self, func_name) for func_name in list(
            methodsWithDecorator(Operation, 'register_non_parallelizable_connector_func'))}

        if include_plugins is None:
            include_plugins = []
        if redis_path is not None:
            self.celerydreamer = CeleryDreamer(include_plugins, redis_path)

    def prepare_results(self):
        prepared_results=[]
        for task_ID, result in self.results.items():
            prepared_results.append({
            'resource_ID' : Task.task_ID_to_resource_ID(task_ID),
            'function_applied' : Task.task_ID_to_function_name(task_ID),
            'time_taken' : self.times.get(task_ID, None),
            'inherited_params' : self.inherited_params.get(task_ID, None),
            'self_params' : self.self_params.get(task_ID, None),
            })
        return prepared_results

    def execute_instructions(self):
        new_nodes_created=[]
        for function_name, params in self.instructions:
            result=self.__getattribute__(function_name)(**params)
            if type(result)==dict and 'new_nodes_created' in result.keys():
                new_nodes_created.extend(result['new_nodes_created'])
        self.instructions=[]
        return new_nodes_created

    def _add_to_instructions(self, params, function_name):
            params['delayed']=False # so that while execution it is actually executed!
            params.pop('self')
            self.instructions.append((function_name, params))


    def add_to_global_cache(self, ID, resource, delayed=True):
        if delayed==True:
            self._add_to_instructions(locals(), inspect.stack()[0][3])
            return

        self.cache[ID]=resource

    def define_template(self, name, processes, delayed=True):
        if delayed==True:
            self._add_to_instructions(locals(), inspect.stack()[0][3])
            return
        self.templates[name]=self._normalize_process_details(processes)

    def add_edge(self, task_ID_A, task_ID_B, silent=False, delayed=True):

        if delayed==True:
            self._add_to_instructions(locals(), inspect.stack()[0][3])
            return

        task_graph = self.task_graph
        task_ID_to_task = self.task_ID_to_task

        if task_ID_A not in task_ID_to_task.keys():
            raise MyException('ERROR: Please, initialize the processes first using the \'add_pipes\' function: '+task_ID_A)
        if task_ID_B not in task_ID_to_task.keys():
            raise MyException('ERROR: Please, initialize the processes first using the \'add_pipes\' function: '+task_ID_B)
        if not silent:
            print("Adding Edge: " + task_ID_A + " --> " + task_ID_B)
        # check if the above breaks the DAG assumptions
        if not nx.has_path(task_graph, task_ID_B, task_ID_A):
            task_graph.add_edge(task_ID_A, task_ID_B)
        else:
            print("LOOP ERROR: :",
                      task_ID_A + " --> " + task_ID_B + " could not be added because it will create a loop!")

    def _normalize_process_details(self, processes):

        # check if there are any calls to templates and unroll them
        processes_ = []
        for process_details in processes:
            if len(process_details) == 1 and process_details[0] in self.templates.keys():
                processes_.extend(self.templates[process_details[0]])
            else:
                processes_.append(process_details)
        processes = processes_

        processes_=[]
        for i, process_details in enumerate(processes):
            if len(process_details)==1: # (process)
                process=process_details[0]
                name=Task.get_process_name(process)+'_'+self.generate_process_ID()
                runtime_params = None
            elif len(process_details)==2:
                if type(process_details[1])==dict: # (process, runtime_params)
                    process = process_details[0]
                    runtime_params = process_details[1]
                    name = Task.get_process_name(process) + '_' + self.generate_process_ID()
                else: # (name, process)
                    name=process_details[0]
                    process = process_details[1]
                    runtime_params = None
            elif len(process_details)==3: # (name, process, runtime_params)
                name = process_details[0]
                process = process_details[1]
                runtime_params = process_details[2]
            else:
                raise MyException('Error: Incorrect number of process details given:', len(process_details))
            processes_.append([name, process, runtime_params])
        return processes_

    def add_pipes(self, resource_ID: str, processes: list, runtime_params_dict: dict = None,
                  resource_dict: dict = None, silent=False, delayed=True):

        """
        processes: a list of tuples (process_name, process)
        runtime_params_dict: {process_name:runtime_params_as_dict}}
        """

        if delayed==True:
            self._add_to_instructions(locals(), inspect.stack()[0][3])
            return

        new_nodes_created=[]

        runtime_params_dict = {} if runtime_params_dict is None else copy.deepcopy(runtime_params_dict)
        resource_dict = {} if resource_dict is None else copy.deepcopy(resource_dict)

        task_graph = self.task_graph
        task_ID_to_task = self.task_ID_to_task

        task_prev = None

        for i, (name, process, runtime_params)  in enumerate(self._normalize_process_details(processes)):
            runtime_params = {} if runtime_params is None else copy.deepcopy(runtime_params)
            op_param_keys = [key for key in runtime_params.keys() if 'op_' in key]
            op_params = {k: runtime_params[k] for k in op_param_keys}
            for k in op_param_keys:
                runtime_params.pop(k)
            task_ID = Task.concoct_task_ID(name, resource_ID)
            # check if it is an invocation request for an internal utility function
            run_function = None
            if type(process) == str and (
                    process in self.connector_functions_mapping.keys() or process in self.non_parallel_connector_functions_mapping.keys()):
                process = self.connector_functions_mapping.get(process, process)
                if type(process) == str:
                    process = self.non_parallel_connector_functions_mapping[process]
                runtime_params['task_ID'] = task_ID
            elif type(process) == str and 'plugins.' in process:
                run_function = process.split('.')[-1]
                process = '.'.join(process.split('.')[:-1])

            if task_ID not in task_ID_to_task.keys():
                # merge the runtime params provided during declaration of the pipe and addition of the pipe!
                runtime_params = merge_dict(runtime_params, runtime_params_dict.get(name, {}))
                task = Task(name, process, resource_ID, self.plugin_collection, self.celerydreamer, run_function,
                            runtime_params, op_params)
                task_ID_to_task[task.get_task_ID()] = task
                if not silent:
                    print('Adding Node:', task_ID)
                task_graph.add_node(task_ID, process=process)
                new_nodes_created.append(task_ID)
            else:
                task = task_ID_to_task[task_ID]
            if i != 0:
                self.add_edge(task_prev.get_task_ID(), task.get_task_ID(), silent=silent)

            # add any explicitly provided resources for this task; these will be made available to any downstream tasks
            for resource_name, resource in resource_dict.get(name, {}).items():
                if task_ID not in self.added_resources[task_ID].keys():
                    self.added_resources[task_ID] = {}
                self.added_resources[task_ID][resource_name] = resource

            # add a new Operation object to the task, this will go into the process as a parameter and can be retrieved from the results
            task.objects['new_operation'] = Operation('new_operation')

            task_prev = task
        return {'new_nodes_created': new_nodes_created}

    def add_connection(self, s_resource_ID, s_name, t_resource_ID, t_name, silent=False, delayed=True):
        """
        To connect two tasks which have been already initialized using the function: add_pipes
        """
        if delayed==True:
            self._add_to_instructions(locals(), inspect.stack()[0][3])
            return

        if s_name in self.templates.keys():
            s_name=self.templates[s_name][-1][0]

        if t_name in self.templates.keys():
            t_name=self.templates[t_name][0][0]

        task_ID_A = Task.concoct_task_ID(s_name, s_resource_ID)
        task_ID_B = Task.concoct_task_ID(t_name, t_resource_ID)
        self.add_edge(task_ID_A, task_ID_B, silent=silent)

    def task_prep(self, task_ID):
        # print('\n')
        # print('DBG: all_added_resources:', self.added_resources.get(task_ID, None))
        # print('DBG: all_inherited_resource:', self.inherited_resource)
        # gather the results from the parent tasks
        params = {}
        resources = {}
        for parent_task_ID in self.task_graph.predecessors(task_ID):
            # print('DBG: parent_task_ID:', parent_task_ID)
            parent_task = self.task_ID_to_task[parent_task_ID]
            for k, v in parent_task.result.items():
                if k not in params.keys():
                    params[k] = []
                    # todo: can provide more options in run_graph as to how to resolve this issue!
                params[k].append(v)

            # add the explicitly added resources of parent tasks to this task's inherited resources
            if parent_task_ID in self.added_resources.keys():
                # print('DBG: found explicit resource of parent')
                if task_ID not in self.inherited_resource.keys():
                    self.inherited_resource[task_ID] = {}
                for resource_name, resource in self.added_resources[parent_task_ID].items():
                    if parent_task_ID not in self.inherited_resource[task_ID].keys():
                        self.inherited_resource[task_ID][parent_task_ID] = {}
                    self.inherited_resource[task_ID][parent_task_ID][resource_name] = resource

            # add the inherited resources of parent as well to this task's inherited resources
            if parent_task_ID in self.inherited_resource.keys():
                # print('DBG: found inherited resource of parent')
                if task_ID not in self.inherited_resource.keys():
                    self.inherited_resource[task_ID] = {}
                for predessesor_task_ID, resource_dict in self.inherited_resource[parent_task_ID].items():
                    for resource_name, resource in resource_dict.items():
                        if predessesor_task_ID not in self.inherited_resource[task_ID].keys():
                            self.inherited_resource[task_ID][predessesor_task_ID] = {}
                        self.inherited_resource[task_ID][predessesor_task_ID][resource_name] = resource

        # collect all resources (current and inherited for this task)
        resources = self.inherited_resource.get(task_ID, {})
        if task_ID in self.added_resources.keys():
            resources[task_ID] = self.added_resources[task_ID]

        # in case of duplicate param names from different task results the results will be in list with the same param name:
        for k, v in params.items():
            if len(v) == 1:
                params[k] = v[0]

        params['op_resources'] = resources
        params['plugin_runner'] = self.plugin_collection

        return params

    def run_graph(self, processes=1, silent=False, no_celery=False):

        if len(self.instructions)!=0:
            print('Executing instructions:')
            self.execute_instructions()

        if processes < 1:
            processes = 1

        start = datetime.datetime.now()
        if processes > 1:
            # todo: make sure redis is running
            # start a single Celery worker that can spawn multiple processes
            # self.celerydreamer.start(concurrency=4)
            if no_celery:
                strigified_list="["+",".join(["\""+item+"\"" for item in self.include_plugins])+"]"
                subprocess.Popen(
                    shlex.split(
                        sys.executable + " -c  'from jpl.pipedreams import celeryapp; cd=celeryapp.CeleryDreamer("+strigified_list+",\""+self.redis_path+"\"); cd.start(concurrency="+str(processes)+")'"),
                    stdout=open(os.devnull, 'wb')
                )

        task_graph = self.task_graph
        next = set()
        added = set()
        completed = set()

        # find all nodes with 0 in-degree and add them to the 'next list'
        seed_tasks = [task_ID for task_ID in task_graph.nodes if task_graph.in_degree(task_ID) == 0]
        next.update(seed_tasks)
        task_completed_count = 0
        pbar = tqdm(total=len(task_graph.nodes))
        while (len(next) != 0 or len(added) != 0):

            # sweep the next queue and add tasks to workers or get them done
            to_remove = []
            to_add = []
            for next_task_ID in next:
                # check if all the parents are in the completed queue
                predecessors = task_graph.predecessors(next_task_ID)
                ripe = True
                for predecessor_ID in predecessors:
                    if predecessor_ID not in completed:
                        ripe = False
                # todo: the parallel and non-parallel execution has a lot of code duplication!
                if ripe:
                    next_task = self.task_ID_to_task[next_task_ID]
                    # gather params from parents
                    params = self.task_prep(next_task_ID)
                    if not silent:
                        print("Adding to Run Queue:", next_task_ID)
                        print("   ---> with inherited params (from parent(s) task result(s)):", params)
                        print("   ---> with self params:", next_task.params)
                    self.inherited_params[next_task_ID]=params
                    self.self_params[next_task_ID] = next_task.params
                    if processes == 1 or next_task_ID in self.non_parallel_connector_functions_mapping.keys():
                        time_task_begin=datetime.datetime.now()
                        next_task.run_task(single_process=True, **params)
                        rs = next_task.result
                        next_task.result = rs['result']
                        self.times[next_task_ID] = rs['time_taken']
                        next_task.postprocess_results()
                        if not silent:
                            print("Non-parallel Task Completed:", next_task_ID)
                            print("     ---> result:", next_task.result)
                        self.results[next_task_ID] = next_task.result
                        pbar.update(1)
                        completed.add(next_task_ID)
                        task_completed_count += 1
                        # see if the task had new instructions within
                        new_instructions=rs['new_operation'].instructions
                        self.instructions=new_instructions
                        new_nodes_created=self.execute_instructions()
                        # find which ones have 0 in degree and attach them to the current task_ID
                        for new_node_created in new_nodes_created:
                            if self.task_graph.in_degree(new_node_created)==0:
                                self.add_edge(next_task_ID, new_node_created, silent=silent, delayed=False)
                        # add the children to next
                        for new_next_task_ID in task_graph.successors(next_task_ID):
                            to_add.append(new_next_task_ID)
                    else:
                        next_task.run_task(single_process=False, **params)
                        added.add(next_task_ID)
                to_remove.append(next_task_ID)
            for task_ID in to_remove:
                next.remove(task_ID)
            for task_ID in to_add:
                next.add(task_ID)

            # sweep the added queue for completed tasks and move them to the completed queue
            to_remove = []
            for added_task_ID in added:
                added_task = self.task_ID_to_task[added_task_ID]
                if str(added_task.result.state) == 'SUCCESS':
                    task_completed_count += 1
                    rs=added_task.result.get()
                    added_task.result = rs['result']
                    self.times[added_task_ID] =rs['time_taken']
                    added_task.postprocess_results()
                    if not silent:
                        print("Parallel Task Completed:", added_task_ID)
                        print("     ---> result:", added_task.result)
                    self.results[added_task_ID] = added_task.result
                    pbar.update(1)
                    completed.add(added_task_ID)
                    to_remove.append(added_task_ID)
                    # see if the task had new instructions within
                    new_instructions = rs['new_operation'].instructions
                    self.instructions = new_instructions
                    new_nodes_created = self.execute_instructions()
                    # find which ones have 0 in degree and attach them to the current task_ID
                    for new_node_created in new_nodes_created:
                        if self.task_graph.in_degree(new_node_created) == 0:
                            self.add_edge(added_task_ID, new_node_created, silent=silent, delayed=False)
                    # add the children to next
                    for next_task_ID in task_graph.successors(added_task_ID):
                        next.add(next_task_ID)

            for task_ID in to_remove:
                added.remove(task_ID)
        pbar.close()
        # kill celery worker
        if processes > 1:
            if no_celery:
                # self.celerydreamer.stop()
                subprocess.call(shlex.split("pkill -f \"celery\""))

        print('num nodes in task graph:', len(task_graph.nodes))
        print('num task completed:', task_completed_count)
        print('time taken:', datetime.datetime.now() - start)



# encoding: utf-8

from .template import Template
import os
import glob

def insensitive_glob(pattern):
    def either(c):
        return '[%s%s]' % (c.lower(), c.upper()) if c.isalpha() else c
    return glob.glob(''.join(map(either, pattern)))


class Disk(Template):
    def __init__(self):
        super().__init__()
        self.description = 'for basic disk file ops'

    def get_bytes(self, path):
        return {'content_bytes': open(path, 'rb').read()}

    def read_str(self, path):
        return {'content_str': open(path, 'r').read()}

    def get_file_size(self, path):
        return int(os.path.getsize(path))

    def isfile(self, path):
        if path.endswith('/'):
            path=path[:-1]
        return os.path.isfile(path)

    def isdir(self, path):
        if path.endswith('/'):
            path=path[:-1]
        return os.path.isdir(path)

    def exists(self, path):
        if path.endswith('/'):
            path=path[:-1]
        return os.path.exists(path)

    def dir_walk(self, path):
        if not path.endswith('/'):
            path=path+'/'
        for root, subdirs, files in os.walk(path):
            subdirs=[os.path.join(path, subdir) for subdir in subdirs]
            files=[os.path.join(path, file) for file in files]
            return subdirs, files
        return [], []

    def recurse(self, path):
        subdirs, files = self.dir_walk(path)
        resource_children = subdirs + files
        for resource in resource_children:
            yield from self.recurse(resource)

        yield path, resource_children

    def download(self, source_path, target_path):
        # TODO: implement a copy operation on disk
        pass

    def search_file(self, path, pattern):
        return [name for name in insensitive_glob(os.path.join(path, pattern))]

    def file_size(self, path):
        return int(os.path.getsize(os.path.join(path)))

    def makedirs(self, path):
        os.makedirs(path, exist_ok=True)



from .template import Template
import pandas as pd


class Txt(Template):
    def __init__(self):
        super().__init__()
        self.description = 'read metadata from txt (tab separated) files'

    def parse_metadata(self, content_bytes, separator='\t'):
        metadata_df = pd.read_csv(content_bytes, sep=separator)
        metadata_df.fillna('', inplace=True)
        return {"metadata": metadata_df} #metadata_df.to_dict('records')}


# encoding: utf-8

'''Plugins ops 🤔'''

import pkgutil
import inspect
from .utils.misc_utils import ignore_unmatched_kwargs


def apply_to_all_methods(decorator):
    # ref: https://stackoverflow.com/questions/6307761/how-to-decorate-all-functions-of-a-class-without-typing-it-over-and-over-for-eac
    def decorate(cls):
        for attr in cls.__dict__: # there's probably a better way to do this
            if callable(getattr(cls, attr)) and attr != "__init__":
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls
    return decorate


class Plugin(object):
    """Base class that each plugin must inherit from. within this class
    you must define the methods that all of your plugins must implement
    """

    @staticmethod
    def kwargs_to_args(func):
        def inner(self, kwargs):
            # print('kwargs_to_args:', func, kwargs)
            return func(self, **kwargs)
        return inner

    def __init__(self):
        self.description = 'UNKNOWN'

        # # NOTE:- IMPORTANT CODE: the below can be applied to all inheriting function statically! removed because we are not dynamically applyin decorators during runtime!

        # for attr in self.__class__.__dict__:
        #     if callable(getattr(self.__class__, attr)) and attr != "__init__":
        #         setattr(self.__class__, attr, ignore_unmatched_kwargs(getattr(self.__class__, attr)))
        #
        # # NOTE:- the following applies the 'kwargs_to_args' decorator to all the methods of this class (even the inherited ones)
        # # this happens because such a child class calls the super init function when initialized.
        # # otherwise @apply_to_all_methods(Plugin.kwargs_to_args) sugar can be used at the level at which the decorator needs to be applied directly
        # for attr in self.__class__.__dict__:
        #     if callable(getattr(self.__class__, attr)) and attr != "__init__":
        #         setattr(self.__class__, attr, Plugin.kwargs_to_args(getattr(self.__class__, attr)))

    def run(self, **kwargs):
        """The method that we expect all plugins to implement. This is the
        method that our framework will call
        """
        # raise NotImplementedError

    def apply(self, func_name, **kwargs):
        """
        Run a specific function by name from the plugin object
        """
        result = ignore_unmatched_kwargs(getattr(self, func_name))(**kwargs)
        return result

def import_module_by_name(module_name):
    plugin_module = __import__(module_name, fromlist=['blah'])
    clsmembers = inspect.getmembers(plugin_module, inspect.isclass)
    for (_, c) in clsmembers:
        # Only add classes that are a sub class of Plugin, but NOT Plugin itself
        if issubclass(c, Plugin) & (c is not Plugin) & (c.__name__ !="Template"):
            # print(f'    Found plugin class: {c.__module__}.{c.__name__}')
            yield c

def find_plugins(package):
    """walk the supplied package to retrieve all plugins (find and initialized objects of a plugin type class)
    """
    imported_package = __import__(package, fromlist=["blah"]) # todo: what is the usage of 'fromlist'!
    plugins=[]
    if hasattr(imported_package, "__path__"):
        for _, pluginname, ispkg in pkgutil.iter_modules(imported_package.__path__, imported_package.__name__ + '.'):
            if not ispkg:
                plugins.extend(list(import_module_by_name(pluginname)))
    else:
        module_name=imported_package.__name__
        plugins.extend(list(import_module_by_name(module_name)))

    return plugins

class PluginCollection(object):

    def __init__(self):
        self.plugin_defs = {}
        self.plugin_inits={}

    def load_plugins(self, name):
        pass

    def get_plugin(self, name):
        if name in self.plugin_inits.keys():
            # print('already init plugin found:', name)
            plugin=self.plugin_inits.get(name)
        else:
            # print('getting a new plugin:', name)
            plugin=self.get_new_plugin(name)
        return plugin

    def get_new_plugin(self, name):
        # look for the plugin in already searched definitions
        plugin_def = self.plugin_defs.get(name, None)
        if plugin_def is not None:
            # print('pre plugin def found, initializing it:', name)
            plugin = plugin_def()  # initialize it
            self.plugin_inits[name] = plugin
        else:
            # search for the plugin
            # print('plugin def not found, looking:', name)
            plugins = find_plugins(name)
            if len(plugins) == 0:
                print('no plugin found:')
                raise ModuleNotFoundError
            elif len(plugins) > 1:
                print('more than one plugin found:', plugins)
                raise ImportError
            else:
                # print('plugin def found, adding and initializing:', name)
                plugin = plugins[0]
                self.plugin_defs[name] = plugin
                plugin = plugin()
                self.plugin_inits[name] = plugin

        return plugin

    def apply(self, template, plugin_name, func_name, kwargs):
        plugin=self.get_plugin(template+'.'+plugin_name)
        result=plugin.apply(func_name, **kwargs)
        return result

"""
refs:
https://www.guidodiepen.nl/2019/02/implementing-a-simple-plugin-framework-in-python/
https://www.datacamp.com/community/tutorials/decorators-python
https://realpython.com/primer-on-python-decorators/#decorators-with-arguments
https://stackoverflow.com/questions/6307761/how-to-decorate-all-functions-of-a-class-without-typing-it-over-and-over-for-eac
https://www.geeksforgeeks.org/creating-decorator-inside-a-class-in-python/
"""



# encoding: utf-8

'''Pipe Dreams'''

import importlib.resources


PACKAGE_NAME = __name__

try:
    __version__ = VERSION = importlib.resources.files(__name__).joinpath('VERSION.txt').read_text().strip()
except AttributeError:
    __version__ = VERSION = importlib.resources.read_text(__name__, 'VERSION.txt')




def add_suffix_to_dict_values(meta_input, suffix):
    return {k: str(v[0])+'_'+suffix for k,v in meta_input.items()}

def retrieve_additional_metadata(metadata, op_resources):
    for parent_task_ID, op_resource in op_resources.items():
        additional_metadata=op_resource['additional_metadata']
        additional_metadata = additional_metadata.to_dict('records')[0]
        for k, v in additional_metadata.items():
            metadata[k]=v
    return metadata



from celery import Celery
from .utils.misc_utils import ignore_unmatched_kwargs
import redis
import datetime

def obj_func_runner(obj, func_name, objects:dict=None, **kwargs):
    if objects is None:
        objects={}
    for object_name, objct in objects.items():
        kwargs[object_name]=objct
    task_begin = datetime.datetime.now()
    result = ignore_unmatched_kwargs(getattr(obj, func_name))(**kwargs)
    time_taken_in_seconds=(datetime.datetime.now()-task_begin).total_seconds()
    to_return= {'result': result, 'time_taken': time_taken_in_seconds}
    for object_name, objct in objects.items():
        to_return[object_name] = kwargs[object_name]
    return to_return

def indi_func_runner(func_object, objects:dict=None, **kwargs):
    if objects is None:
        objects={}
    for object_name, objct in objects.items():
        kwargs[object_name]=objct
    task_begin = datetime.datetime.now()
    result = ignore_unmatched_kwargs(func_object)(**kwargs)
    time_taken_in_seconds=(datetime.datetime.now()-task_begin).total_seconds()
    to_return = {'result': result, 'time_taken': time_taken_in_seconds}
    for object_name, objct in objects.items():
        to_return[object_name] = kwargs[object_name]
    return to_return

class CeleryDreamer():

    def __init__(self, plugin_list, redis_url=None):
        if redis_url[-1]!='/':
            redis_url+='/'
        # flush redis
        r = redis.Redis.from_url(redis_url)
        r.flushdb()
        self.BROKER_URL = redis_url+'0'
        self.BACKEND_URL = redis_url+'1'
        self.app = Celery('proj',
                          broker=self.BROKER_URL,
                          backend=self.BACKEND_URL,
                          include=plugin_list)
        # Optional configuration, see the application user guide.
        self.app.conf.update(
            result_expires=3600,
            serializer='pickle',
            result_serializer='pickle',
            task_serializer='pickle',
            accept_content=['pickle', 'json'],
            result_accept_content=['pickle', 'json']
        )
        self.celery_obj_func_runner = self.app.task(obj_func_runner)
        self.celery_indi_func_runner = self.app.task(indi_func_runner)

    def start(self, concurrency):
        self.app.control.purge()
        self.worker = self.app.Worker(concurrency=concurrency, loglevel="DEBUG")
        self.worker.start()

    def stop(self):
        self.worker.stop()




# encoding: utf-8

from .template import Template
import configparser

class MyException(Exception):
    pass

class Cfg(Template):
    def __init__(self):
        super().__init__()
        self.description = 'read metadata from .cfg files'

    def parse_metadata(self, content_str):
        metadata = {}
        config = configparser.RawConfigParser()
        config.optionxform = str
        config.read_string(content_str)
        header = config.sections()
        if len(header) == 0 or header[0] not in ['Collection', 'Dataset', "File"]:
            raise MyException(
                'Improper formatting of the metadata file! Provide the correct header/section; Collection, Dataset or File!')
        else:
            header = header[0]
        for key in config[header]:
            value = config[header][key].strip()
            key = key.strip()
            value = [item.strip() for item in value.split('|') if item.strip() != '']
            metadata[key] = value
        return {"metadata": metadata, "header": header}



# encoding: utf-8

from jpl.pipedreams.plugins_ops import Plugin


class Template(Plugin):
    def __init__(self):
        super().__init__()
        self.description = 'for reading different metadata files'

    def parse_metadata(self, *args, **kwargs):
        raise NotImplementedError



from .template import Template
import pandas as pd

class Excel(Template):
    def __init__(self):
        super().__init__()
        self.description = 'read metadata from excel files'

    def parse_metadata(self, content_bytes):
        xl = pd.ExcelFile(content_bytes)
        metadata_df = pd.read_excel(content_bytes, sheet_name=xl.sheet_names[0])
        metadata_df.fillna('', inplace=True)
        return {"metadata": metadata_df} # metadata_df.to_dict('records')}


# encoding: utf-8

from jpl.pipedreams.plugins_ops import Plugin


class Template(Plugin):
    def __init__(self):
        super().__init__()
        self.description = 'for basic file ops'

    def get_bytes(self, *args, **kwargs):
        raise NotImplementedError

    def read_str(self, *args, **kwargs):
        raise NotImplementedError

    def get_file_size(self, *args, **kwargs):
        raise NotImplementedError

    def isfile(self, *args, **kwargs):
        raise NotImplementedError

    def isdir(self, *args, **kwargs):
        raise NotImplementedError

    def exists(self, *args, **kwargs):
        raise NotImplementedError

    def dir_walk(self, *args, **kwargs):
        raise NotImplementedError

    def download(self, *args, **kwargs):
        raise NotImplementedError

    def search_file(self, *args, **kwargs):
        raise NotImplementedError



# encoding: utf-8

'''Demonstration. This is a demo to show the usage of jpl.pipedreams'''


import os
import time
from jpl.pipedreams.data_pipe import Operation
from plugins.general.for_demo import add_suffix_to_dict_values, retrieve_additional_metadata

# Let us declare some global variables first
base_data_path = os.path.join(os.path.dirname(__file__), 'data') # the path to the test data

# here we will state the url for the redis instance we seek to use
redis_path='redis://localhost:6379'
# you will also have to provide the relative path to your directory with all your plugins.
# This helps us find all your functions and register them as Celery Workers.
plugins_dir_list=['plugins']


# Hello World!
# ============
#
# Here we will go through the steps rather quickly and explain things more clearly in the next section.
#
# First, initialize an operation (an operation contains a task graph associated with it):
my_first_operation = Operation(name='my_first_operation', redis_path=redis_path, include_plugins=plugins_dir_list)

# Declare your process using a plugin function:
process_list = [
    ["hello_world_read", "plugins.file_ops.disk.read_str", {"path": os.path.join(base_data_path, 'mydata0.txt')}]
]

"""
- The 'process_list' may contain multiple process declarations with the following format:
    (<process_name: give a name to this process>, <plugin path or a function etc.>, <any params needed for the function to run>)
"""
# add the declaration to your operation
my_first_operation.add_pipes("mydata0.txt", process_list)
# run the operation
my_first_operation.run_graph()


# ========================= The one where things start to get complicated...I mean complex!
# ======================================================================================================================
"""
- Here we will create a pipeline where a .cfg file is read and some simple process is applied to the contents.
"""
my_second_operation = Operation(name='my_second_operation', redis_path=redis_path, include_plugins=plugins_dir_list)
# declare the processes that run sequentially in a single template; we call this a template cause you can reuse it later.
main_task_template = [
    # read the metadata file from disk
    ["read_cfg", "plugins.file_ops.disk.read_str", None],
    # parse the string into a dictionary
    ["parse_cfg", "plugins.metadata_parsers.cfg.parse_metadata", None],
    # change the name to match the input of the next function
    ["result_change_name", 'change_name', {'in_to_out': {'metadata': 'meta_input'}}],
    # do some simple processing to the metadata
    ["process_metadata", add_suffix_to_dict_values, {"suffix": "|", "op_out": ["metadata"]}],
]

# using your template, initialize/add tasks to your operation all at once
my_second_operation.add_pipes(resource_ID="mydata1.cfg", processes=main_task_template, runtime_params_dict={'read_cfg': {'path': os.path.join(base_data_path, 'mydata1.cfg')}})


"""
- The above 'add_pipes' call, takes each item in the declared template and initializes a unique task with 'resource_ID' and 
    'processe_name' together as the primary key.
- Notice that we can use any imported function, like 'add_suffix_to_dict_values' 
    or, a well defined plugin (using our interface), like 'jpl.pipedreams.plugins.file_ops.disk.read_str'
    or, even some internal connector functions, like 'change_name'
- Using the 'runtime_params_dict' you can also pass more params for you task functions. As a rule of thumb, 
    you can add the params that can be reused, in the template declaration, and the ones that may change with each task 
    initialization, during the 'add_pipes' call.
- Notice that the internal function 'change_name' makes sure that the 'metadata' param is fed into the 'meta_input' 
    param of the function 'add_suffix_to_dict_values'.
- Notice the 'op_out' param in the last process, what it does is makes sure that the result of 
    the 'add_suffix_to_dict_values' function gets the name 'metadata'.
"""
my_second_operation.run_graph()
print('Results:', my_second_operation.results)


# ========================= The one where we do much less work!
# ======================================================================================================================
my_second_operation.add_pipes(resource_ID="mydata2.cfg", processes=main_task_template, runtime_params_dict={'read_cfg': {'path': os.path.join(base_data_path, 'mydata2.cfg')}})

"""
- Here we used our earlier template to initialize some more tasks but with a different 'resource_ID' 
    (making them independent tasks of course) and we also feed a different file to this pipeline
- Notice that we used the filename also as the 'resource_ID'. To be honest, you can use any name 
    above and it won't be an issue.
"""
my_second_operation.run_graph(processes=1)
"""
- If you give the 'process' a value more than 1, it uses Celery and Redis to parallelize the graph execution.
- Try it but make sure you have a redis client running and have Celery installed!
"""
time.sleep(3)


# ========================= The one where the long lost twins meet!
# ======================================================================================================================
# declare a new process, to be used to merge results from the already declared pipelines
merge_template = ("data_merge", 'merge', {"in_multi": "metadata",  "merge_type": "collect"})
my_second_operation.add_pipes(resource_ID="data_merge1", processes=[merge_template])
"""
- Here we are intending to merge the two results from earlier, both called 'metadata', such that the duplicate keys 
    with different values get joined together in a list. This highlights the usefulness of another internal function.
"""
my_second_operation.add_connection(s_resource_ID='mydata1.cfg', s_name='process_metadata', t_resource_ID='data_merge1', t_name='data_merge')
my_second_operation.add_connection(s_resource_ID='mydata2.cfg', s_name='process_metadata', t_resource_ID='data_merge1', t_name='data_merge')
"""
- Since we had already declared and initialized all the tasks, in above we just connect the tasks together
"""
my_second_operation.run_graph()


# ========================== The one where the kids steal their parent's car keys
# ======================================================================================================================
my_third_operation = Operation(name='my_third_operation', redis_path=redis_path, include_plugins=plugins_dir_list)

add_op_resource = [
    # read the bytes of the file from disk
    ["read_excel", "plugins.file_ops.disk.get_bytes", None],
    # parse the excel into a dataframe
    ["excel_to_df", "plugins.metadata_parsers.excel.parse_metadata", None],
    # add the result to the op resources
    ["add_extra_metedata_resource", 'add_to_op_resource', {"in_to_op": {'metadata': 'additional_metadata'}}],
    # remove the op resource part of the result from the result
    ["remove_resource_from_result", 'remove_keys', {"keys_to_remove":["metadata"]}]
]
"""
- We defined a process template where we intend to read an excel sheet and make it available to all our successor tasks
    by using the internal function 'add_to_op_resource'.
- Notice that we also removed it from the output of the task, as it was completely unnecessary since the very next task 
    may not need it!
"""
main_task_template = [
    # read the metadata file from disk
    ["read_cfg", "plugins.file_ops.disk.read_str", None],
    # parse the string into a dictionary
    ["parse_cfg", "plugins.metadata_parsers.cfg.parse_metadata", None],
    # add the additional metadata to the main metadata
    ["use_the_added_resource", retrieve_additional_metadata, None]
]
"""
- Here the function 'retrieve_additional_metadata' gives a good example of how to use the op_resource we added earlier.
    Go chek out the implementation of it.
- This helps us load resources only one time and let it be used by inheriting tasks!
"""
runtime_params_dict = {
    'read_cfg': {'path': os.path.join(base_data_path, 'mydata1.cfg')},
    'read_excel': {'path': os.path.join(base_data_path, 'mymapping.xlsx')}
}
my_third_operation.add_pipes(resource_ID="mydata1.cfg", processes=add_op_resource+main_task_template, runtime_params_dict=runtime_params_dict)
my_third_operation.run_graph()

"""
Now, let us simplify things further by defining templates.
"""
my_fourth_operation = Operation(name='my_fourth_operation', redis_path=redis_path, include_plugins=plugins_dir_list)

main_task_template = [
    # read the metadata file from disk
    ["read_cfg", "plugins.file_ops.disk.read_str"],
    # parse the string into a dictionary
    ["plugins.metadata_parsers.cfg.parse_metadata"],
    # change the name to match the input of the next function
    ['change_name', {'in_to_out': {'metadata': 'meta_input'}}],
    # do some simple processing to the metadata
    [add_suffix_to_dict_values, {"suffix": "|", "op_out": ["metadata"]}],
]
# add this to the operation as a template that can be used by its name later.
my_fourth_operation.define_template('metadata_reader', main_task_template)

my_fourth_operation.add_pipes(resource_ID="mydata1.cfg", processes=[['metadata_reader']], runtime_params_dict={'read_cfg': {'path': os.path.join(base_data_path, 'mydata1.cfg')}})
my_fourth_operation.add_pipes(resource_ID="mydata2.cfg", processes=[['metadata_reader']], runtime_params_dict={'read_cfg': {'path': os.path.join(base_data_path, 'mydata2.cfg')}})
my_fourth_operation.add_pipes(resource_ID="data_merge1", processes=[["data_merge", 'merge', {"in_multi": "metadata",  "merge_type": "collect"}]])
my_fourth_operation.add_connection(s_resource_ID='mydata1.cfg', s_name='metadata_reader', t_resource_ID='data_merge1', t_name='data_merge')
my_fourth_operation.add_connection(s_resource_ID='mydata2.cfg', s_name='metadata_reader', t_resource_ID='data_merge1', t_name='data_merge')
my_fourth_operation.run_graph()

"""
let us create a graph within a graph!
"""

my_fifth_operation = Operation(name='my_fifth_operation', redis_path=redis_path, include_plugins=plugins_dir_list)

main_task_template = [
    # read the metadata file from disk
    ["read_cfg", "plugins.file_ops.disk.read_str"],
    # parse the string into a dictionary
    ["plugins.metadata_parsers.cfg.parse_metadata"],
    # change the name to match the input of the next function
    ['change_name', {'in_to_out': {'metadata': 'meta_input'}}],
    # do some simple processing to the metadata
    [add_suffix_to_dict_values, {"suffix": "|", "op_out": ["metadata"]}],
]
# add this to the operation as a template that can be used by its name later.
my_fifth_operation.define_template('metadata_reader', main_task_template)

def some_process(new_operation):
    new_operation.add_pipes(resource_ID="mydata1.cfg", processes=[['metadata_reader']], runtime_params_dict={'read_cfg': {'path': os.path.join(base_data_path, 'mydata1.cfg')}})
    return

my_fifth_operation.add_pipes(resource_ID="graph_create_1", processes=[[some_process]])
my_fifth_operation.run_graph()




import functools
import inspect
import datetime

class MyException(Exception):
    pass


def collect_dict(dict_a, dict_b):
    """
    collect everything from dict_a and dict_b in this manner:
    dict_a={A:B, C:D}
    dict_b={A:E}

    return {A:[B, E], C:D}
    """

    dict_return={}
    intersect=set(dict_a.keys()).intersection(set(dict_b.keys()))
    for key in intersect:
        if type(dict_a[key])!=list:
            dict_a[key]=[dict_a[key]]
        if type(dict_b[key])!=list:
            dict_b[key]=[dict_b[key]]

        dict_return[key]=list(set(dict_a[key]+dict_b[key])) # merge and deduplicate

    for key in set(dict_a.keys())-intersect:
        dict_return[key]=dict_a[key]

    for key in set(dict_b.keys())-intersect:
        dict_return[key]=dict_b[key]

    return dict_return

def collect_dicts(list_of_dict):

    if type(list_of_dict)!=list:
        list_of_dict=[list_of_dict]

    combined_dict= {}
    for i, dict in enumerate(list_of_dict):
        if i==0:
            combined_dict=dict
            continue
        combined_dict=collect_dict(combined_dict, dict)

    return combined_dict

def merge_dict(dict_a, dict_b):
    """
    _b replaces _b for any duplicate keys
    """
    merged_dict={}
    for k, v in dict_a.items():
        merged_dict[k]=v
    for k, v in dict_b.items():
        merged_dict[k]=v
    return merged_dict

def merge_dicts(list_of_dict):

    if type(list_of_dict)!=list:
        list_of_dict=[list_of_dict]

    merged_dict= {}
    for i, dict in enumerate(list_of_dict):
        if i==0:
            merged_dict=dict
            continue
        merged_dict=merge_dict(merged_dict, dict)

    return merged_dict

def remove_keys(dict_, list_of_keys, **kwargs):
    for k in list_of_keys:
        if k in dict_.keys():
            dict_.pop(k)
    return dict_

def boolit(str):
    if str in ['True', 'true']:
        return True
    else:
        return False

def ignore_unmatched_kwargs(f):

    """
    ref: https://stackoverflow.com/questions/26515595/how-does-one-ignore-unexpected-keyword-arguments-passed-to-a-function
    Make function ignore unmatched kwargs.

    If the function already has the catch all **kwargs, do nothing.
    """
    if any(param.kind == inspect.Parameter.VAR_KEYWORD for param in inspect.signature(f).parameters.values()):
        return f

    #
    @functools.wraps(f)
    def inner(*args, **kwargs):
        # For each keyword arguments recognised by f,
        # take their binding from **kwargs received
        filtered_kwargs = {
            name: kwargs[name]
            for name, param in inspect.signature(f).parameters.items() if (
                                                                                  param.kind is inspect.Parameter.KEYWORD_ONLY or
                                                                                  param.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD
                                                                          ) and
                                                                          name in kwargs
        }
        return f(*args, **filtered_kwargs)

    return inner

def nested_set(dic, keys, value):
    if len(keys)==0:
        dic=value
        return dic
    for key in keys[:-1]:
        dic = dic.setdefault(key, {})
    dic[keys[-1]] = value
    return dic
